import struct
from dataclasses import dataclass
from enum import IntEnum
from io import BufferedIOBase
from numpy.typing import NDArray
import numpy as np


class ColourSpace(IntEnum):
    """Enum to differentiate colour spaces."""

    sRGB = 0
    Linear = 1


class FrameType(IntEnum):
    """Enum to differentiate frame types."""

    # Key Frame types are independently decodeable or encodable.
    Key = 0
    # Predicted Frame types are encoded based on the previous key frame.
    Predicted = 1


@dataclass
class QovHeader:
    """Data type to facilitate reading and writing the QOV file header."""

    magic: str = "qoiv"
    width: int = 640
    height: int = 480
    colourspace: ColourSpace = ColourSpace.sRGB
    # There are 3 padding bytes after the colourspace field to align the structure to 16 bytes.

    @staticmethod
    def read(file: BufferedIOBase) -> "QovHeader":
        """Read the header from the provided file handle."""
        header_packed: bytes = file.read(16)
        if len(header_packed) != 16:
            raise ValueError(f"Invalid header size, was {len(header_packed)}")
        magic, width, height, colourspace = struct.unpack("<4sIIBxxx", header_packed)
        if magic != b"qoiv":
            raise ValueError("Invalid magic number")
        if colourspace not in ColourSpace:
            raise ValueError("Invalid colourspace")
        return QovHeader(
            magic=magic.decode("utf-8"),
            width=width,
            height=height,
            colourspace=ColourSpace(colourspace),
        )

    def write(self, file: BufferedIOBase) -> None:
        """Write the header to the provided file handle."""
        if self.magic != "qoiv":
            raise ValueError("Invalid magic number")
        if self.colourspace not in ColourSpace:
            raise ValueError("Invalid colourspace")
        file.write(
            struct.pack(
                "<4sIIBxxx",
                self.magic.encode("utf-8"),
                self.width,
                self.height,
                self.colourspace,
            )
        )


class PixelHashMap:
    """Hash map for constant time lookup of previously used colours."""

    def __init__(self, size: int = 64):
        """Construct a fixed size hash map for colours."""
        self.size = size
        self.pixels = np.array([[0, 0, 0]] * self.size)

    def push(self, pixel: NDArray[np.uint8]) -> int:
        """Push a pixel into the hash map."""
        r, g, b = pixel

        index = self.index_of(r, g, b)
        self.pixels[index] = pixel
        return index

    def __getitem__(self, index: int) -> NDArray[np.uint8]:
        """Get an item from the hash map by index."""
        if index < 0 or index >= self.size:
            raise IndexError("Index out of bounds")
        return self.pixels[index]

    def __contains__(self, pixel: NDArray[np.uint8]) -> bool:
        """Check the colour is in the map"""
        r, g, b = pixel
        return PixelHashMap.pixel_equal(self.pixels[self.index_of(r, g, b)], pixel)

    def clear(self):
        """Clear the hash map."""
        self.pixels.fill(0)

    def index_of(self, r: int, g: int, b: int) -> int:
        """Calculate the index of a pixel in the hash map."""
        return (r * 3 + g * 5 + b * 7) % self.size

    @staticmethod
    def pixel_equal(a: NDArray[np.uint8], b: NDArray[np.uint8]) -> bool:
        """Check if two pixels are equal."""
        return a[0] == b[0] and a[1] == b[1] and a[2] == b[2]


@dataclass
class QovFrameHeader:
    """Header for a single frame."""

    frame_type: FrameType

    @staticmethod
    def read(file: BufferedIOBase) -> "QovFrameHeader":
        """Read the frame header from the provided file handle."""
        frame_type = FrameType(int.from_bytes(file.read(1)))
        if frame_type not in FrameType:
            raise ValueError("Invalid frame type")
        return QovFrameHeader(frame_type=frame_type)

    def write(self, file: BufferedIOBase):
        """Write the frame header to the provided file handle."""
        file.write(struct.pack("<B", self.frame_type))



from io import BytesIO
from numpy.typing import NDArray
import numpy as np
from .types import QovHeader, QovFrameHeader, FrameType, PixelHashMap
from .opcodes import (
    RgbOpcode,
    DiffOpcode,
    IndexOpcode,
    FrameRunOpcode,
    RunOpcode,
    DiffFrameOpcode,
)


class Decoder:
    """Decode a QOIV file into frames."""

    def __init__(self, file: BytesIO):
        """Construct a new decoder."""
        self.file = file
        self.header = QovHeader.read(file)
        self.first_frame_pos = file.tell()
        self.pixel_count = self.header.width * self.header.height
        self.key_pixels = PixelHashMap()

    def __iter__(self) -> "Decoder":
        """Setup the iterator"""
        self.frame_pos = self.first_frame_pos
        self.file.seek(self.frame_pos)
        self.key_pixels = PixelHashMap()
        self.key_frame_flat = None
        return self

    def __next__(self):
        """Get the next frame."""
        return self.read_frame()

    def read_frame(self) -> NDArray[np.uint8]:
        """Read the next frame from the file."""
        frame_header = QovFrameHeader.read(self.file)

        frame = np.zeros((self.header.height, self.header.width, 3), dtype=np.uint8)

        pixels = PixelHashMap()

        pixel_read = 0
        while pixel_read < self.pixel_count:
            cy = pixel_read // self.header.width
            cx = pixel_read % self.header.width
            if RgbOpcode.is_next(self.file):
                opcode = RgbOpcode.read(self.file)
                frame[
                    cy,
                    cx,
                ] = [opcode.r, opcode.g, opcode.b]
                pixels.push(frame[cy, cx])
                pixel_read += 1
            elif DiffOpcode.is_next(self.file):
                opcode = DiffOpcode.read(self.file)
                frame[cy, cx] = frame[
                    (pixel_read - 1) // self.header.width,
                    (pixel_read - 1) % self.header.width,
                ] + np.array([opcode.dr, opcode.dg, opcode.db])
                pixels.push(frame[cy, cx])
                pixel_read += 1
            elif RunOpcode.is_next(self.file):
                opcode = RunOpcode.read(self.file)
                last_pixel = frame[
                    (pixel_read - 1) // self.header.width,
                    (pixel_read - 1) % self.header.width,
                ]
                frame.reshape(-1, 3, copy=False)[
                    pixel_read : pixel_read + opcode.run
                ] = last_pixel
                pixel_read += opcode.run
            elif IndexOpcode.is_next(self.file):
                index_opcode = IndexOpcode.read(self.file)
                frame[cy, cx] = pixels[index_opcode.index]
                pixel_read += 1
            elif DiffFrameOpcode.is_next(self.file):
                if self.key_frame_flat is None:
                    raise ValueError("Unexpected DiffFrameOpcode without key frame.")
                if frame_header.frame_type == FrameType.Key:
                    raise ValueError("Unexpected DiffFrameOpcode in key frame.")

                diff_frame_opcode = DiffFrameOpcode.read(self.file)
                if diff_frame_opcode.key_frame:
                    if diff_frame_opcode.use_index:
                        pixel = self.key_pixels[diff_frame_opcode.index] + np.array(
                            [
                                diff_frame_opcode.dr,
                                diff_frame_opcode.dg,
                                diff_frame_opcode.db,
                            ]
                        )

                        frame[cy, cx] = pixel
                        pixels.push(pixel)
                    else:
                        pixel = self.key_frame_flat[pixel_read] + np.array(
                            [
                                diff_frame_opcode.diff + diff_frame_opcode.dr,
                                diff_frame_opcode.diff + diff_frame_opcode.dg,
                                diff_frame_opcode.diff + diff_frame_opcode.db,
                            ]
                        )

                        frame[cy, cx] = pixel
                        pixels.push(pixel)
                else:
                    raise NotImplementedError()
                pixel_read += 1
            elif FrameRunOpcode.is_next(self.file) and self.key_frame_flat is not None:
                opcode = FrameRunOpcode.read(self.file)
                if not opcode.is_keyframe:
                    raise NotImplementedError()

                frame.reshape(-1, 3, copy=False)[
                    pixel_read : pixel_read + opcode.run
                ] = self.key_frame_flat[pixel_read : pixel_read + opcode.run]
                pixel_read += opcode.run
            else:
                raise ValueError("Unexpected opcode in key frame.")

        if frame_header.frame_type == FrameType.Key:
            self.key_pixels = pixels
            self.key_frame_flat = frame.reshape(-1, 3)

        return frame



from .types import ColourSpace, QovHeader, PixelHashMap, QovFrameHeader, FrameType
from .opcodes import (
    Opcode,
    RgbOpcode,
    IndexOpcode,
    DiffOpcode,
    RunOpcode,
    DiffFrameOpcode,
    FrameRunOpcode,
)
from typing import Optional
import numpy as np
from numpy.typing import NDArray
from dataclasses import dataclass
from typing import List
from io import BufferedIOBase


@dataclass
class EncodedFrame:
    """A helper class to represent how to encode a frame as a series of opcodes."""

    header: QovFrameHeader
    opcodes: List[Opcode]

    def __len__(self):
        """Report the size of the frame in bytes."""
        return sum([len(opcode) for opcode in self.opcodes])

    def write(self, file: BufferedIOBase) -> None:
        """Convert and write the frame to the provided file handle."""
        self.header.write(file)
        for opcode in self.opcodes:
            opcode.write(file)


class Encoder:
    """This class is responsible for encoding frames into the QOV format."""

    def __init__(
        self,
        file: BufferedIOBase,
        width: int,
        height: int,
        colourspace: ColourSpace,
        keyframe_interval: Optional[int] = None,
    ):
        """Construct a new encoder."""
        self.header = QovHeader(width=width, height=height, colourspace=colourspace)
        self.file = file
        self.keyframe_interval = keyframe_interval
        self.last_keyframe: Optional[NDArray[np.uint8]] = None
        self.frames_since_last_keyframe: int = -1
        self.total_frames = 0
        self.header.write(file)
        self.pixels = PixelHashMap()

    def __repr__(self) -> str:
        """String representation of the encoder."""
        return f"Encoder(width={self.header.width}, height={self.header.height}, colourspace={self.header.colourspace}, keyframe_interval={self.keyframe_interval}, total_frames={self.total_frames})"

    def trigger_keyframe(self) -> None:
        """Ensure that the next frame is a keyframe."""
        self.frames_since_last_keyframe = -1

    @property
    def is_next_frame_keyframe(self) -> bool:
        """Determine if the next frame is a keyframe."""
        if self.frames_since_last_keyframe == -1:
            self.frames_since_last_keyframe = 0
            return True

        if self.keyframe_interval is None:
            return True

        if self.frames_since_last_keyframe >= self.keyframe_interval:
            self.frames_since_last_keyframe = 0
            return True

        return False

    def encode_keyframe(
        self, frame: NDArray[np.uint8], pixels: PixelHashMap
    ) -> EncodedFrame:
        """Encode a frame as a keyframe."""
        return self._encode_frame(frame, pixels, None, None)

    @staticmethod
    def pixel_equal(a: NDArray[np.uint8], b: NDArray[np.uint8]) -> bool:
        """Check if two pixels are equal."""
        return a[0] == b[0] and a[1] == b[1] and a[2] == b[2]

    @staticmethod
    def pixels_equal(pixels: NDArray[np.uint8]) -> int:
        equal_rows = np.all(pixels[1:] == pixels[0], axis=1)
        return (
            int(np.argmax(~equal_rows)) if not np.all(equal_rows) else len(equal_rows)
        )

    def _encode_frame(
        self,
        frame: NDArray[np.uint8],
        pixels: PixelHashMap,
        key_frame_flat: Optional[NDArray[np.uint8]],
        key_pixels: Optional[PixelHashMap],
    ) -> EncodedFrame:
        """Encode a single frame."""
        opcodes: List[Opcode] = []

        last_pixel: Optional[NDArray[np.uint8]] = None
        last_pixel_count: int = 0
        is_kf_flat = key_frame_flat is not None
        is_kf_pixels = key_pixels is not None
        frame_flat = frame.reshape(-1, 3, copy=False)
        pixel_pos = 0
        pixel_len = len(frame_flat)
        while pixel_pos < pixel_len:
            pixel = frame_flat[pixel_pos]
            if last_pixel is not None:
                # Handle runs
                count = Encoder.pixels_equal(frame_flat[pixel_pos - 1 : pixel_pos + 62])
                if count > 0:
                    opcodes.append(RunOpcode(run=count))
                    pixel_pos += count
                    continue

                dr, dg, db = pixel - last_pixel
                if -2 <= dr < 2 and -2 <= dg < 2 and -2 <= db < 2:
                    opcodes.append(DiffOpcode(dr, dg, db))
                    pixels.push(pixel)
                    last_pixel = pixel
                    pixel_pos += 1
                    continue

            if pixel in pixels:
                opcodes.append(
                    IndexOpcode(index=pixels.index_of(pixel[0], pixel[1], pixel[2]))
                )
                pixels.push(pixel)
                last_pixel = pixel
                pixel_pos += 1
                continue

            if is_kf_flat:
                count = Encoder.get_pixels_equal(
                    frame_flat[pixel_pos : pixel_pos + 128],
                    key_frame_flat[pixel_pos : pixel_pos + 128],
                )
                match count:
                    case 0:
                        pass
                    case 1:
                        opcodes.append(DiffFrameOpcode(True, False, 0, 0, 0, diff=0))
                        pixels.push(pixel)
                        last_pixel = pixel
                        pixel_pos += 1
                        continue
                    case _:
                        m = min(128, count)
                        opcodes.append(FrameRunOpcode(is_keyframe=True, run=m))
                        pixel_pos += m
                        last_pixel = frame_flat[pixel_pos - 1]
                        continue

            if is_kf_pixels and pixel in key_pixels:
                key_index = key_pixels.index_of(pixel[0], pixel[1], pixel[2])
                opcodes.append(DiffFrameOpcode(True, True, 0, 0, 0, index=key_index))
                pixels.push(pixel)
                last_pixel = pixel
                pixel_pos += 1
                continue

            # Fallback to RGB opcode
            pixels.push(pixel)
            last_pixel = pixel
            pixel_pos += 1
            opcodes.append(RgbOpcode(r=pixel[0], g=pixel[1], b=pixel[2]))

        if last_pixel_count > 0:
            opcodes.append(RunOpcode(run=last_pixel_count))

        frame_type = FrameType.Key if key_pixels is None else FrameType.Predicted
        return EncodedFrame(
            header=QovFrameHeader(
                frame_type=frame_type,
            ),
            opcodes=opcodes,
        )

    @staticmethod
    def get_pixels_equal(a: NDArray[np.uint8], b: NDArray[np.uint8]) -> int:
        """Return the number of pixels that are equal in two arrays."""
        equal_rows = np.all(a == b, axis=1)
        return (
            int(np.argmax(~equal_rows)) if not np.all(equal_rows) else len(equal_rows)
        )

    def encode_predicted(
        self,
        frame: NDArray[np.uint8],
        pixels: PixelHashMap,
        key_frame_flat: NDArray[np.uint8],
        key_pixels: PixelHashMap,
    ) -> EncodedFrame:
        """Encode a predicted frame."""
        return self._encode_frame(frame, pixels, key_frame_flat, key_pixels)

    def push(self, frame: NDArray[np.uint8]) -> None:
        """Push a new frame into the encoder."""

        if self.is_next_frame_keyframe:
            self.pixels.clear()
            encoded = self.encode_keyframe(frame, self.pixels)
            self.key_frame_flat = frame.reshape(-1, 3)
            encoded.write(self.file)

        else:
            encoded = self.encode_predicted(
                frame, PixelHashMap(), self.key_frame_flat, self.pixels
            )
            encoded.write(self.file)
            self.frames_since_last_keyframe += 1

        self.total_frames += 1

    def flush(self) -> None:
        """Flush the encoder to the file."""
        self.file.flush()



import struct
import os
from collections.abc import Sized
from typing import Protocol, Optional
from io import BufferedIOBase
from dataclasses import dataclass


class Opcode(Sized, Protocol):
    """An interface for opcodes used in the QOV encoding."""

    def write(self, file: BufferedIOBase) -> None:
        """Write the opcode to file."""
        ...


@dataclass
class RgbOpcode(Opcode):
    """The QOI_OP_RGB opcode, encodes a single RGB pixel."""

    r: int
    g: int
    b: int

    def write(self, file: BufferedIOBase) -> None:
        """Write the RGB opcode to the provided file handle."""
        file.write(struct.pack("<B3B", 0xFE, self.r, self.g, self.b))

    def __len__(self) -> int:
        """Fixed size of 4"""
        return 4

    @staticmethod
    def is_next(file: BufferedIOBase) -> bool:
        """Read the next byte and determine if it is a RgbOpcode."""
        code = file.read(1)
        file.seek(-1, os.SEEK_CUR)
        return code == b"\xfe"

    @staticmethod
    def read(file: BufferedIOBase) -> "RgbOpcode":
        """Read an RGB opcode from the provided file handle."""
        code = file.read(4)
        if len(code) != 4 or code[0] != 0xFE:
            raise ValueError("Invalid RGB opcode")
        r, g, b = struct.unpack("<3B", code[1:])
        return RgbOpcode(r, g, b)


@dataclass
class IndexOpcode(Opcode):
    """The QOI_OP_INDEX opcode, encodes an index into the pixel hash map."""

    index: int

    def write(self, file: BufferedIOBase) -> None:
        """Write the Index opcode to the provided file handle."""
        if 0 > self.index or self.index > 63:
            raise ValueError("Index must be between 0 and 63")
        file.write(struct.pack("<B", self.index & 0x3F))

    def __len__(self) -> int:
        """Fixed size of 1"""
        return 1

    @staticmethod
    def is_next(file: BufferedIOBase) -> bool:
        """Read the next byte and determine if it is an IndexOpcode."""
        code = file.read(1)
        file.seek(-1, os.SEEK_CUR)
        return 0 <= code[0] <= 63

    @staticmethod
    def read(file: BufferedIOBase) -> "IndexOpcode":
        """Read an Index opcode from the provided file handle."""
        code = file.read(1)
        if len(code) != 1 or not (0 <= code[0] <= 63):
            raise ValueError("Invalid Index opcode")
        return IndexOpcode(index=code[0])


@dataclass
class DiffOpcode(Opcode):
    """The QOI_OP_DIFF opcode, encodes a difference between the last pixel and the current pixel."""

    dr: int
    dg: int
    db: int

    def write(self, file: BufferedIOBase) -> None:
        """Write the Diff opcode to the provided file handle."""
        if not (-2 <= self.dr < 2 and -2 <= self.dg < 2 and -2 <= self.db < 2):
            raise ValueError("Diff values must be between -2 and 1")
        file.write(
            struct.pack(
                "<B", 0x40 | (self.dr + 2) << 4 | (self.dg + 2) << 2 | (self.db + 2)
            )
        )

    def __len__(self) -> int:
        """Fixed size of 1"""
        return 1

    @staticmethod
    def is_next(file: BufferedIOBase) -> bool:
        """Determine if the next opcode is a DiffOpcode."""
        code = file.read(1)
        file.seek(-1, os.SEEK_CUR)
        return code[0] & 0xC0 == 0x40

    @staticmethod
    def read(file: BufferedIOBase) -> "DiffOpcode":
        """Read a Diff opcode from the provided file handle."""
        code = file.read(1)
        if len(code) != 1 or code[0] & 0xC0 != 0x40:
            raise ValueError("Invalid Diff opcode")
        dr = ((code[0] >> 4) & 0x03) - 2
        dg = ((code[0] >> 2) & 0x03) - 2
        db = (code[0] & 0x03) - 2
        return DiffOpcode(dr, dg, db)


@dataclass
class RunOpcode(Opcode):
    """The QOI_OP_RUN opcode, encodes a run of identical pixels."""

    run: int

    def write(self, file: BufferedIOBase) -> None:
        """Write the Run opcode to the provided file handle."""
        if not (1 <= self.run <= 62):
            raise ValueError("Run value must be between 1 and 62")
        file.write(struct.pack("<B", 0xC0 | (self.run - 1)))

    def __len__(self) -> int:
        """Fixed size of 1"""
        return 1

    @staticmethod
    def is_next(file: BufferedIOBase) -> bool:
        """Determine if the next opcode is a RunOpcode."""
        code = file.read(1)
        file.seek(-1, os.SEEK_CUR)
        return (code[0] & 0xC0) == 0xC0 and code[0] != 0xFF and code[0] != 0xFE

    @staticmethod
    def read(file: BufferedIOBase) -> "RunOpcode":
        """Read a Run opcode from the provided file handle."""
        code = file.read(1)
        if len(code) != 1 or code[0] & 0xC0 != 0xC0:
            raise ValueError("Invalid Run opcode")
        run = (code[0] & 0x3F) + 1
        return RunOpcode(run=run)


@dataclass
class FrameRunOpcode(Opcode):
    """This Opcode replaces the QOI_OP_RGBA opcode, and is used to represent a run of identical pixels from a previous frame."""

    is_keyframe: bool
    run: int

    def write(self, file: BufferedIOBase) -> None:
        """Write the FrameRunOpcode to the provided file handle."""
        if not (1 <= self.run <= 128):
            raise ValueError("Run value must be between 1 and 128")
        file.write(struct.pack("<BB", 0xFF, (self.run - 1) | self.is_keyframe << 7))

    def __len__(self) -> int:
        return 2

    @staticmethod
    def is_next(file: BufferedIOBase) -> bool:
        code = file.read(1)
        file.seek(-1, os.SEEK_CUR)
        return code[0] == 0xFF

    @staticmethod
    def read(file: BufferedIOBase) -> "FrameRunOpcode":
        """Read a FrameRunOpcode from the provided file handle."""
        code = file.read(2)
        if len(code) != 2 or code[0] != 0xFF:
            raise ValueError("Invalid FrameRun opcode")
        is_keyframe = (code[1] & 0x80) != 0
        run = (code[1] & 0x7F) + 1
        return FrameRunOpcode(is_keyframe, run)


class DiffFrameOpcode(Opcode):
    """This Opcode replaces the QOI_OP_LUMA opcode, and is used to query data from previous frames.

    If key_frame is True, then the value is fetched from the keyframe.
    If key_frame is False, then the value is fetched from the previous frame.
    If use_index is True, then the value is fetched from the PixelHashMap, for the keyframe, using the index.
    If use_index is False, then the value is fetched from the frame buffer.

    key_frame == False and use_index == True is an invalid state.

    dr, dg, and db are the differences in red, green, and blue channels respectively from the pixel fetched from the above location.
    If use_index is False, these differences are added to the value in index, which will be between -32..31.
    """

    def __init__(
        self,
        key_frame: bool,
        use_index: bool,
        dr: int,
        dg: int,
        db: int,
        index: Optional[int] = None,
        diff: Optional[int] = None,
    ):
        """Construct a new DiffFrameOpcode."""
        self.key_frame = key_frame
        self.use_index = use_index
        if diff and index:
            raise ValueError("Only one of index or diff can be provided")
        if index is not None:
            self.index = index
        elif diff is not None:
            self.index = diff + 32
        else:
            raise ValueError("Either index or diff must be provided")
        self.dr = dr
        self.dg = dg
        self.db = db

    def __len__(self) -> int:
        """Fixed size of 2"""
        return 2

    def __eq__(self, other: object) -> bool:
        """Check for equality with another DiffFrameOpcode."""
        if isinstance(other, DiffFrameOpcode):
            return (
                self.key_frame == other.key_frame
                and self.use_index == other.use_index
                and self.index == other.index
                and self.dr == other.dr
                and self.dg == other.dg
                and self.db == other.db
            )
        return False

    @staticmethod
    def is_next(file: BufferedIOBase) -> bool:
        """Determine if the next opcode is a DiffFrameOpcode."""
        code = file.read(1)
        file.seek(-1, os.SEEK_CUR)
        return (code[0] & 0xC0) == 0x80

    @staticmethod
    def read(file: BufferedIOBase) -> "DiffFrameOpcode":
        """Read a DiffFrameOpcode from the provided file handle."""
        b = file.read(2)
        if len(b) != 2 or (b[0] & 0xC0) != 0x80:
            raise ValueError("Invalid DiffFrame opcode")
        key_frame = (b[1] & 0x80) != 0
        use_index = (b[1] & 0x40) != 0
        index = b[0] & 0x3F
        dr = ((b[1] >> 4) & 0x03) - 2
        dg = ((b[1] >> 2) & 0x03) - 2
        db = (b[1] & 0x03) - 2
        return DiffFrameOpcode(key_frame, use_index, dr, dg, db, index)

    @property
    def index(self) -> int:
        """Getter for index."""
        return self._index

    @property
    def diff(self) -> int:
        """calculate the difference from the base index of 32."""
        return self.index - 32

    @index.setter
    def index(self, value: int):
        """Set the index."""
        if value < 0 or value > 63:
            raise ValueError("Index must be between 0 and 63")
        self._index = value

    @property
    def dr(self) -> int:
        """Getter for dr (red difference)."""
        return self._dr

    @dr.setter
    def dr(self, value: int):
        """Set the red difference."""
        if not (-2 <= value < 2):
            raise ValueError("dr must be between -2 and 1")
        self._dr = value

    @property
    def dg(self) -> int:
        """Getter for dg (green difference)."""
        return self._dg

    @dg.setter
    def dg(self, value: int):
        """Set the green difference."""
        if not (-2 <= value < 2):
            raise ValueError("dg must be between -2 and 1")
        self._dg = value

    @property
    def db(self) -> int:
        """Getter for db (blue difference)."""
        return self._db

    @db.setter
    def db(self, value: int):
        """Set the blue difference."""
        if not (-2 <= value < 2):
            raise ValueError("db must be between -2 and 1")
        self._db = value

    def write(self, file: BufferedIOBase) -> None:
        """Write the DiffFrameOpcode to the provided file handle."""
        file.write(
            struct.pack(
                "<BB",
                0x80 | self.index,
                int(self.key_frame) << 7
                | int(self.use_index) << 6
                | (self.dr + 2) << 4
                | (self.dg + 2) << 2
                | (self.db + 2),
            )
        )



from numpy.typing import NDArray
import typer
from pathlib import Path
from pyqoiv.encode import Encoder
from pyqoiv.types import ColourSpace
import ffmpeg
import numpy as np
import tqdm as tqdm
from typing import Generator

app = typer.Typer()


@app.command()
def encode(input_file: Path, output_file: Path) -> None:
    """Encode a qoiv formatted file from any video file ffmpeg supports."""
    probe = ffmpeg.probe(str(input_file))
    video_stream = next(
        (stream for stream in probe["streams"] if stream["codec_type"] == "video"), None
    )
    if video_stream is None:
        raise typer.BadParameter("Input file does not contain a video stream.")
    width = int(video_stream["width"])
    height = int(video_stream["height"])
    frame_rate = int(video_stream["avg_frame_rate"].split("/")[0])
    duration = float(probe["format"]["duration"])
    approx_frames = int(frame_rate * duration)

    encoder = Encoder(
        output_file.open("wb"), width, height, ColourSpace.Linear, keyframe_interval=20
    )

    out = (
        ffmpeg.input(str(input_file))
        .output("pipe:", format="rawvideo", pix_fmt="rgb24")
        .run_async(pipe_stdout=True, quiet=True)
    )

    def read_frames() -> Generator[NDArray[np.uint8], None, None]:
        while True:
            in_bytes = out.stdout.read(width * height * 3)
            if not in_bytes:
                break
            in_frame = np.frombuffer(in_bytes, np.uint8).reshape((height, width, 3))
            yield in_frame

    for frame in tqdm.tqdm(read_frames(), total=approx_frames, desc="Encoding"):
        encoder.push(frame)
    out.stdout.close()

    pass


@app.command()
def decode() -> None:
    """Decode qoiv formatted file"""
    pass
